from armet.connectors.flask import resources as flask_resources
from armet.connectors.sqlalchemy import resources as sqlalchemy_resources
from armet import utils
from alchemist import db
import warnings

__all__ = [
    'Resource',
    'ModelResource',
]


class Resource(flask_resources.Resource):

    @property
    def session(self):
        return db.session

    def route(self, *args, **kwargs):
        try:
            # Continue on with the cycle.
            result = utils.super(Resource, self).route(*args, **kwargs)

            # Commit the session.
            db.commit()

            # Return the result.
            return result

        except:
            # Something occurred; rollback the session.
            db.rollback()

            # Re-raise the exception.
            raise

        finally:
            # Remove our reference to the locally-aquired session.
            db.remove()


class ModelResource(sqlalchemy_resources.ModelResource):

    def route(self, *args, **kwargs):
        return utils.super(sqlalchemy_resources.ModelResource, self).route(
            *args, **kwargs)



# -*- coding: utf-8 -*-
from sqlalchemy import orm
from sqlalchemy.ext.declarative.api import DeclarativeMeta


class Query(orm.Query):
    """Base query; dervied from the SQLAlchemy query object.
    """

    def __new__(cls, entities=None, *args, **kwargs):
        # Find and instantiate the query class defined on the entity.
        entity = entities[-1] if entities else None
        if isinstance(entity, DeclarativeMeta):
            if hasattr(entity, '__query__'):
                # Found a query class; set it.
                cls = entity.__query__

        # Continue as normal.
        return super().__new__(cls)



# -*- coding: utf-8 -*-
import sys
from termcolor import colored

# Alias some colors.
colored_info = lambda x: colored(x, 'white', attrs=['dark'])
colored_command = lambda x: colored(x, 'cyan')
colored_name = lambda x: colored(x, 'white')


def print_command(info, command, name, *more):
    print(colored_info(info),
          colored_command(command),
          colored_name(name),
          colored_info(' '.join(more)),
          file=sys.stderr)



# -*- coding: utf-8 -*-
import pytest
import flask
from alchemist import db, commands, application
from wsgi_intercept import add_wsgi_intercept, remove_wsgi_intercept
from wsgi_intercept.requests_intercept import install_opener, uninstall_opener
import threading
import requests


class TestBase:

    #! List of packages used during testing; limits the number of packages
    #! that are reset after each test run.
    #! Default: reset all packages.
    packages = None

    #! Host at which the test server will bein intercepted at.
    host = 'localhost'

    #! Port at which the test server will being intercepted at.
    port = 5000

    @pytest.fixture(autouse=True, scope='class')
    def server(cls, request):
        # Install the WSGI interception layer.
        install_opener()

        # Enable the WSGI interception layer.
        add_wsgi_intercept(cls.host, cls.port, lambda: flask.current_app)

        # Add a finalizer to remove the interception layer.
        def finalize():
            remove_wsgi_intercept(cls.host, cls.port)
            uninstall_opener()

        request.addfinalizer(finalize)

    @pytest.fixture(autouse=True, scope='class')
    def database(cls, request):
        # Initialize the database access layer.
        commands.db.clear(names=cls.packages, echo=False)
        commands.db.init(names=cls.packages, echo=False)

        # TODO: Load any required fixtures.

    def setup_class(cls):
        # Create a shortcut for querying because we're all lazy and we
        # know it.
        cls.Q = lambda s, x: db.session.query(x)

    def teardown(self):
        # Rollback the session.
        db.rollback()

        # Flush the database access layer.
        commands.db.flush(names=self.packages, echo=False)

        # Commit all of the deletes.
        db.commit()

        # TODO: Re-load any desired fixtures.

    def request(self, path='', url=None, *args, **kwargs):
        # Helper to forward the port and host if url is not specified.
        if url is None:
            url = 'http://{}:{}'.format(self.host, self.port)

        # Set some defaults for requests.
        kwargs.setdefault('allow_redirects', False)

       # Forward to requests.
        with application.app_context():
            response = requests.request(*args, url=url + path, **kwargs)

        # Refresh the current session (remove its transaction lock so
        # changes will be seen).
        db.commit()

        # Return the response.
        return response

    def head(self, *args, **kwargs):
        kwargs.setdefault('method', 'HEAD')
        return self.request(*args, **kwargs)

    def options(self, *args, **kwargs):
        kwargs.setdefault('method', 'OPTIONS')
        return self.request(*args, **kwargs)

    def get(self, *args, **kwargs):
        kwargs.setdefault('method', 'GET')
        return self.request(*args, **kwargs)

    def post(self, *args, **kwargs):
        kwargs.setdefault('method', 'POST')
        return self.request(*args, **kwargs)

    def put(self, *args, **kwargs):
        kwargs.setdefault('method', 'PUT')
        return self.request(*args, **kwargs)

    def patch(self, *args, **kwargs):
        kwargs.setdefault('method', 'PATCH')
        return self.request(*args, **kwargs)

    def delete(self, *args, **kwargs):
        kwargs.setdefault('method', 'DELETE')
        return self.request(*args, **kwargs)



# -*- coding: utf-8 -*-
import flask
import sys
import types
import os
from collections import Mapping, OrderedDict
import sqlalchemy as sa
from sqlalchemy.ext.declarative import DeclarativeMeta
from importlib import import_module
from flask import current_app as application
from .meta import version as __version__, description as __doc__  # NOQA
from . import utils

__all__ = [
    'Alchemist',
    'application',
    'configured'
]


# Create signal namespace.
signals = flask.signals.Namespace()

# Create named signals.
configured = signals.signal('configured')


def wrap_module(name, application):

    # Wrap the module with a nested module that
    # contains property accessors for late-bound configuration.
    class Module(types.ModuleType):

        @property
        def application(self):
            application.configure()
            return application

    # Update the inner module with the actual contents.
    instance = Module(name)
    instance.__dict__.update(sys.modules[name].__dict__)

    # Return the wrapped module.
    return instance


class Alchemist(flask.Flask):

    def __init__(self, name, *args, **kwargs):
        # Initialize the flask application context.
        super().__init__(name, *args, **kwargs)

        # Replace the package with a container module that contains
        # a property reference to the application.
        sys.modules[name] = wrap_module(name, self)

        #! Collections of database models and metadata.
        self.models = OrderedDict()
        self.metadata = OrderedDict()

    def _apply_site_configuration(self):
        # Gather configuration from the following places (with precedence):
        #   1. $<package.application.name>_SETTINGS_MODULE
        #   2. $ALCHEMIST_SETTINGS_MODULE
        #   3. <project>.settings

        # Set default settings.
        # The default packages array contains just the alchemist lib which
        # includes all default commands.
        self.config['PACKAGES'] = ['alchemist']

        # The default server configuration enables threading.
        self.config['SERVER'] = {
            'host': 'localhost',
            'port': 8000,
            'threading': True
        }

        try:
            # Attempt to get configuration from the project settings.
            self.config.from_object('{}.settings'.format(self.name))

        except ImportError:
            # No settings module or package.
            pass

        # Attempt to get configuration from the environment variables.
        for name in ['ALCHEMIST', self.name.upper().replace('.', '_')]:
            var = '%s_SETTINGS_MODULE' % name
            if var in os.environ:
                self.config.from_envvar(var)

    def configure(self):
        """Collect settings and configure the application instance."""

        # Establish an application context.
        context = self.app_context()
        context.push()

        # Detect if we are being invoked by a test runner.
        # Checks the first and second arguments to determine if we are being
        # run by a test runner.
        # Able to match `py.test`, `nosetests`, `alchemist test`,
        # and `python setup.py test`.
        self.config['TESTING'] = False
        count = 3 if os.path.basename(sys.argv[0]).startswith('python') else 2
        for argument in sys.argv[0:count]:
            if 'test' in argument:
                # We are, in fact, being run by a test runner.
                self.testing = self.config['TESTING'] = True
                break

        # Apply the initial site configuration.
        self._apply_site_configuration()

        # After the initial configuration is gathered; each registered
        # package is searched for a settings module or package and that
        # is loaded then the site configuration is re-applied (to
        # keep precedence and allow dynamic behavior).
        all_models = set()
        for name in self.config.get('PACKAGES', ()):
            try:
                # Attempt to get configuration from the settings module.
                self.config.from_object('{}.settings'.format(name))

            except ImportError:
                # No settings module.
                pass

            try:
                # Attempt to import a models module or name.
                models = import_module('{}.models'.format(name))

            except ImportError:
                # No models module found.
                models = None

            # Import the name itself.
            package = import_module(name)

            # Initialize models set.
            self.models[name] = set()

            # Create filterer that only grabs instances of DeclarativeMeta.
            restrict = lambda o: set(filter(
                lambda x: isinstance(x, DeclarativeMeta),
                o._decl_class_registry.values()))

            # Try searching through the namespaces of both the package
            # and the models module to find the metadata.
            for module in package, models:
                if module:
                    for cls in vars(module).values():
                        if isinstance(cls, type) and cls not in all_models:
                            meta = getattr(cls, 'metadata', None)
                            if meta and isinstance(meta, sa.MetaData):
                                # Found a match; move along.
                                self.metadata[name] = meta

                                # Update registry.
                                models = restrict(cls)
                                self.models[name].update(models)
                                all_models.update(models)

                                # Clear modules.
                                modules, package = None, None
                                break

            # Re-apply the site configuration.
            self._apply_site_configuration()

        # Process and resolve database configuration.
        self.databases = {}
        for name, db in self.config.get('DATABASES', {}).items():
            if isinstance(db, Mapping):
                # Build the database URI.
                uri = utils.build_database_uri(testing=self.testing, **db)

            else:
                # The database object should be the URI.
                uri = db

            # Create the database engine.
            self.databases[name] = sa.create_engine(
                uri, **db.get('options', {}))

        # Send the configured signal.
        configured.send(self)

        # Release the application context.
        context.pop()



# -*- coding: utf-8 -*-
import sys
from flask.ext import script
from termcolor import colored
from alchemist import application, db


def _make_context(quiet=False):
    # Start by adding the session instance to the namespace.
    namespace = {'db': db, 'session': db.session}

    if not quiet:
        # Let the user know.
        print(colored('from alchemist import db', 'green'))
        print(colored('session = db.session', 'green'))

    # Iterate through all models and add those to the namespace.
    module = {}
    for name, models in application.models.items():
        for model in models:
            if model.__module__ not in module:
                module[model.__module__] = []

            module[model.__module__].append(model.__name__)
            namespace[model.__name__] = model

    if not quiet:
        # Let the user know.
        for name, models in module.items():
            items = ', '.join(models)
            print(colored('from %s import %s' % (name, items), 'green'))

    # Return the constructed namespace.
    return namespace


class Shell(script.Shell):
    """
    Runs a Python shell inside an application context with all
    models pre-loaded.
    """

    def __init__(self, *args, **kwargs):
        # Default the context maker.
        kwargs.setdefault('make_context', _make_context)

        # Continue initialization.
        super().__init__(*args, **kwargs)

    def get_options(self):
        options = super().get_options()

        # Add an additional option that allows for piping of input
        # to the shell.
        options += script.Option('--pipe', action="store_true", dest='pipe'),

        return options

    def run(self, pipe, *args, **kwargs):
        if pipe:
            # User is attempting to pipe in script through stdin.
            text = sys.stdin.read()
            exec(text, None, _make_context(quiet=True))
            return

        # Fallback to normal cycle.
        super().run(*args, **kwargs)



# -*- coding: utf-8 -*-
from .db import Database
from .show import Show
from .shell import Shell
from .load import Load
from .server import Server

__all__ = [
    'Database',
    'Show',
    'Shell',
    'Load',
    'Server'
]



# -*- coding: utf-8 -*-
from flask.ext import script
from alchemist.conf import settings


class Server(script.Server):
    """Runs the application server for development use.
    """

    name = 'run'

    def __init__(self, *args, **kwargs):
        # Pull server configuration out of configuration.
        server = settings.get('SERVER', {})
        server['host'] = server.get('host', 'localhost')
        server['port'] = server.get('port', 8000)
        server['threaded'] = server.get('threaded', True)

        # Default the server configuration for flask-script.
        kwargs.update(server)

        # Continue initialization.
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
import flask
import collections

__all__ = [
    'settings'
]


class Settings(collections.Mapping):
    """
    A proxy dictionary-like object that proxies to the current app
    configuration; or, if not in an active application context, pretends
    to be an empty dictionary.
    """

    def __getitem__(self, name):
        try:
            return flask.current_app.config[name]

        except RuntimeError:
            raise KeyError

    def __contains__(self, name):
        try:
            return name in flask.current_app.config

        except RuntimeError:
            return False

    def __len__(self):
        try:
            return len(flask.current_app.config)

        except RuntimeError:
            return 0

    def __iter__(self):
        try:
            return iter(flask.current_app.config)

        except:
            return iter(())

# Instantiate lazy-bound settings object.
settings = Settings()



# -*- coding: utf-8 -*-
import contextlib
import datetime
from sqlalchemy import schema
from sqlalchemy.orm.query import Query
from flask.ext import script
from alchemist import application
from .utils import print_command
from sqlalchemy_utils import render_statement


def limit_metadata(names):
    # Iterate through all registered packages if one is not specified.
    metadata = application.metadata
    if names:
        # Are the names registered?
        metadata = {k: v for k, v in metadata.items() if k in names}
        if not metadata:
            raise ValueError("One of the listed packages is not registered.")

    # Return metadata to use.
    return metadata


def init(names=None, sql=False, echo=True):
    """Synchronize the database; update all specified packages' tables."""
    # TODO: Support multi-db routing.
    engine = application.databases['default']

    # Iterate through each collected package metadata.
    for name, metadata in limit_metadata(names).items():
        if echo:
            # Log the sequence.
            print_command('alchemist db', 'init', name)

        # Iterate through all tables.
        for table in metadata.sorted_tables:
            if not table.exists(engine):
                if echo:
                    # Log the sequence.
                    print_command('alchemist db', 'create', table.name,
                                  'default')

                if sql:
                    # Print the creation statement.
                    print(str(schema.CreateTable(table)) + ';')

                else:
                    # Create the table.
                    table.create(engine)


def clear(names=None, sql=False, echo=True):
    """Clear the database; drop all specified tables."""
    # TODO: Support multi-db routing.
    engine = application.databases['default']

    # Iterate through each collected package metadata.
    for name, metadata in reversed(list(limit_metadata(names).items())):
        if echo:
            # Log the sequence.
            print_command('alchemist db', 'clear', name)

        # Iterate through all tables.
        for table in reversed(metadata.sorted_tables):
            if table.exists(engine):
                if echo:
                    # Log the sequence.
                    print_command('alchemist db', 'drop', table.name,
                                  'default')

                if sql:
                    # Print the creation statement.
                    print(str(schema.DropTable(table)) + ';')

                else:
                    # Create the table.
                    table.drop(engine)


def flush(names=None, sql=False, echo=True):
    """Flush the database; delete data from specified tables."""
    # TODO: Support multi-db routing.
    engine = application.databases['default']

    # Iterate through each collected package metadata.
    for name, metadata in reversed(list(limit_metadata(names).items())):
        if echo:
            # Log the sequence.
            print_command('alchemist db', 'flush', name)

        # Iterate through all tables; deleting those neccessary.
        for table in reversed(metadata.sorted_tables):
            if table.exists(engine):
                # Create the DELETE statement.
                statement = table.delete()

                if echo:
                    # Log the sequence.
                    print_command('alchemist db', 'shell', statement,
                                  'default')

                if sql:
                    # Log the statement.
                    print(render_statement(statement, bind=engine) + ';')

                else:
                    with contextlib.closing(engine.connect()) as connection:
                        # Execute the statement.
                        connection.execute(statement)


def show(names=None):
    """List all discovered models."""
    # Iterate through each collected package metadata.
    for name, metadata in limit_metadata(names).items():
        # Log the sequence.
        print_command('alchemist db', 'show', name)

        # Iterate through all tables.
        for table in metadata.sorted_tables:
            # Log the sequence.
            print_command('alchemist db', 'show table', table.name)


class Database(script.Manager):

    #! Name of the command as it is invoked on the command line.
    name = 'db'

    def __init__(self, *args, **kwargs):
        # Default the usage description.
        kwargs.setdefault('usage', 'Perform various database operations.')

        # Continue the initialization.
        super().__init__(*args, **kwargs)

        # Declare common option arguments.
        name_kwargs = dict(
            dest='names', nargs='*',
            help='The package to perform the operation on; '
                 'defaults to all.')

        # Add commands.
        for command in [init, clear, flush]:
            # Add the name option to limit operation scope.
            command = self.option(**name_kwargs)(command)

            # Add the SQL option to print SQL instead.
            command = self.option(
                '-S', '--sql', dest='sql', required=False, default=False,
                action='store_true',
                help='Print the SQL statements to stdout instead '
                     'executing them directly.')(command)

        # Add the show command.
        self.option(**name_kwargs)(show)



# -*- coding: utf-8 -*-
import weakref
import re
import sqlalchemy as sa
from sqlalchemy.orm import object_session
from sqlalchemy.orm.util import has_identity
from sqlalchemy.ext.declarative import declared_attr, DeclarativeMeta, base
from alchemist.conf import settings


def _package_of(module):
    """Gets the root package name of the passed `models` package.
    """

    # Get the registered package this model belongs to.
    package = module.split('.')
    while package:
        # Is this name a registered package?
        test = '.'.join(package)
        if test in settings.get('PACKAGES', []):
            # This is the package we are in.
            return test

        # Remove the right-most segment.
        package.pop()

    if not package:
        # No package was found to be registered; attempt to guess the
        # right package name; strip all occurrances of '.models' from the
        # pacakge name.
        return module.replace('.models', '')


#! Dictionary of metadata classes; keyed by package.
_model_metadata = {}


#! Dictionary of class registries; keyed by package.
_model_registry = {}


class ModelBase(DeclarativeMeta):

    @property
    def _decl_class_registry(self):
        return getattr(self, '_registry', None)

    @property
    def query(self):
        """Create an object session and return the query object."""
        return __import__('alchemist.db').db.session.query(self)

    def __new__(cls, name, bases, attrs):
        # Don't process further if this is the base.
        for base in bases:
            if isinstance(base, ModelBase):
                # This is twice derived.
                break

        # Nope; this is the base.
        else:
            return super().__new__(cls, name, bases, attrs)

        # Check for existing metadata.
        package = _package_of(attrs['__module__'])
        if package not in _model_registry:
            _model_registry[package] = weakref.WeakValueDictionary()
            _model_metadata[package] = sa.MetaData()

        # Set new registry.
        attrs['_registry'] = _model_registry[package]

        # Add metadata and registry to the attributes.
        attrs['metadata'] = _model_metadata[package]

        # Continue processing.
        return super().__new__(cls, name, bases, attrs)


class Model(metaclass=ModelBase):
    """Declares the base model.

    This provides various helpers, defaults, and utilities for
    sqlalchemy-derived models.
    """

    __abstract__ = True

    __init__ = base._declarative_constructor

    @declared_attr
    def __tablename__(cls):
        """
        Underscorizes the class name combined with the pacakge name in order
        to form a normal name for a table in SQL.
        """
        package = _package_of(cls.__module__).lower()
        name = '%s.%s' % (package, cls.__name__.lower())
        name = re.sub(r'([A-Z])', r'_\1', name)
        name = re.sub(r'\.+', r'_', name)
        return name

    def save(self, session=None, commit=True):
        """
        Save the changes to the model. If the model has not been persisted
        then it adds the model to the declared session. Then it flushes the
        object session and optionally commits it.

        @param[in] session
            A specific session to use instead of the thread-local, scoped
            session.
        """
        if has_identity(self):
            # Object has already been persisted to the database; grab its
            # session.
            session = object_session(self)

        else:
            # Ensure we have a database session.
            if session is None:
                session = __import__('alchemist.db').db.session

            # Object has not been persisted to the database.
            session.add(self)

        if commit:
            # Commit the session as requested.
            session.commit()

        else:
            # Just flush the session; do not commit.
            session.flush()



# -*- coding: utf-8 -*-
from flask.ext.script import Command


class Show(Command):
    """Show the name and package of the application.
    """

    def run(self):
        print(__import__('alchemist').application)



# -*- coding: utf-8 -*-
import os
from collections import defaultdict
from flask.ext.script import Command, Option
from alchemist import db
from alchemist.db.models import _package_of
from .utils import print_command
from pkgutil import get_importer
from sqlalchemy import event
from sqlalchemy.orm import Mapper


class Load(Command):
    """Loads the passed named fixture or file.
    """

    # The idea is to eventually extend this command to support:

    # @code
    # $ alchemist load <file.(json|yml|py)> [--database <database>]
    # $ alchemist load <name> [--database <database>]
    # @endcode

    # Where the second form would traverse each package and look for a
    # <name>.(json|yml|py) in the package or in a fixtures folder of the
    # package and load each one.

    option_list = [Option(dest='filename')]

    def run(self, filename):
        # Resolve the absoulte path.
        path = os.path.abspath(filename)

        # Resolve the module name.
        name = os.path.basename(os.path.splitext(path)[0])

        # Create a after_* hook so that we can record what happens.
        models = defaultdict(int)

        @event.listens_for(Mapper, 'after_insert')
        def hook(mapper, connection, target):
            target_cls = type(target)
            package = _package_of(target_cls.__module__)
            name = '%s.%s' % (package, target_cls.__name__)
            models[name] += 1

        try:
            # Import and execute the file.
            imp = get_importer(os.path.dirname(filename))
            imp.find_module(name).load_module(name)

            # Commit the session.
            db.commit()

        except:
            # Something happened; rollback the session.
            db.rollback()

            # Re-raise so the console gets the traceback.
            raise

        finally:
            # Close the session.
            db.close()

        # Get sizes for logging.
        max_count = len(str(max(models.values())))

        # Let the user know.
        for name, count in models.items():
            print_command('alchemist db', 'insert',
                          ('{:>%s} {}' % (max_count)).format(count, name),
                          'default')



 # -*- coding: utf-8 -*-
from collections import defaultdict
from flask.ext import script
from flask.ext.script import Option
import os
import sys
import pkgutil
import glob
import flask
from importlib import import_module
import colorama
from termcolor import colored


def _is_package():
    """Gets if the current directory is a package."""
    # Save the basename (possible package name).
    cwd = os.getcwd()
    name = os.path.basename(cwd)

    # Go back one directory.
    os.chdir('..')

    # Check if any of the found modules match the possible package name.
    result = False
    for _, module_name, is_pkg in pkgutil.iter_modules('.'):
        if module_name == name and is_pkg:
            result = True
            break

    # Return to the previous directory.
    os.chdir(name)
    return result


def discover():
    """
    Discovers the application context to use from traversing the package
    directory structure.

    The algorithm works as follows:
        - If currently in a python package; find the root package.
            > Use the root package `.app` or `.application` object.
        - Else, check all modules in current and sub-directories
          for `.app` or `.application`.
    """

    name = None
    while _is_package():
        # We are in a python package. Iterate back until we are at the
        # root of it.
        name = os.path.basename(os.getcwd())
        os.chdir('..')

    if name is None:
        # Not in a python package; iterate through all modules in the
        # current directory and one-level below each.
        for paths in ['.', (glob.glob('src') + glob.glob('*'))]:
            for imp, module_name, _ in pkgutil.iter_modules(paths):
                if module_name == 'setup':
                    # Don't look at setup.py
                    continue

                m = imp.find_module(module_name).load_module()
                app = getattr(m, 'application', getattr(m, 'app', None))
                if isinstance(app, flask.Flask):
                    # Found an application object.
                    # Add it to the system path.
                    sys.path.append(os.getcwd())

                    # Re-import the module.
                    m = import_module(module_name)

                    # Return the object.
                    return getattr(m, 'application', getattr(m, 'app', None))

        # Still found nothing; give up.
        return None

    # Import the package.
    package = pkgutil.get_importer(os.getcwd()).find_module(name).load_module()
    app = getattr(package, 'application', getattr(package, 'app', None))
    if isinstance(app, flask.Flask):
        # Found an application object.
        # Add it to the system path.
        sys.path.append(os.getcwd())

        # Re-import the module.
        package = import_module(name)

        # Return the object.
        return getattr(package, 'application', getattr(package, 'app', None))


class Command(script.Command):

    #! The defined name to be invoked.
    name = None

    #! The defined namespace of the command.
    #! This is a flexible sub-manager.
    namespace = None

    def log(self, *args):
        name = self.namespace if self.namespace else self.name
        name = 'alchemist %s' % name

        print(colored(name, 'white', attrs=['dark']), *args, file=sys.stderr)

    def warn(self, *args):
        self.log(colored('WARN', 'yellow', attrs=['bold', 'reverse']), *args)

    def fail(self, *args):
        self.log(colored('FAIL', 'red', attrs=['bold', 'reverse']), *args)
        sys.exit(1)


class Manager(script.Manager):

    def __init__(self, *args, **kwargs):
        # Attempt to discover an application.
        app = discover()

        # Set default arguments.
        kwargs.setdefault('with_default_commands', False)

        # Continue initialization.
        super().__init__(app or flask.Flask('alchemist'), *args, **kwargs)

        # Initialize command list.
        self.__commands = []

        # Establish an application context (if we can).
        context = None
        if app is not None:
            context = app.app_context()
            context.push()

        # Add commands from all registered packages.
        packages = app.config.get('PACKAGES', []) if app else ['alchemist']
        for package in packages:
            try:
                # Attempt to load a `.commands` module from the package.
                module = import_module('.commands', package=package)

            except ImportError:
                # No command package.
                continue

            # Iterate through all exposed modules and register anything
            # that looks like a command.
            for name in getattr(module, '__all__', vars(module).keys()):
                obj = getattr(module, name, None)
                if isinstance(obj, type):
                    if (issubclass(obj, script.Manager)
                            or issubclass(obj, script.Command)):
                        # Add the exposed command.
                        self.add_command(obj)

        if context:
            # Release the application context.
            context.pop()

        # Initialize cross-platform terminal colors.
        colorama.init()

    def add_command(self, name, command=None):
        # Allow a 1-argument form of this command where the name
        # is found on the command object as `.name`.
        if command is None:
            command, name = name, None

        # Grab the `.name` on the command or the `__name__` of the
        # command object if not passed a name.
        if name is None:
            name = getattr(command, 'name', command.__name__.lower())

        # Instantiate the command if we were given a type.
        if isinstance(command, type):
            command = command()

        # Add commands to command list.
        self.__commands.append((name, command))

    def handle(self, *args, **kwargs):
        if self.__commands:
            # Resolve command list.
            namespaced_commands = defaultdict(lambda: [])
            for name, command in self.__commands:
                namespace = getattr(command, 'namespace', None)
                if namespace:
                    namespaced_commands[namespace].append((name, command))
                    continue
                super().add_command(name, command)

            # Clear command list.
            self.__commands = []

            # Resolve namespaced commands.
            for namespace in namespaced_commands:
                manager = script.Manager()
                for name, command in namespaced_commands[namespace]:
                    manager.add_command(name, command)
                super().add_command(namespace, manager)

        # Continue on with handling the command.
        super().handle(*args, **kwargs)

    def run(self, *args, **kwargs):
        # Ensure we have an established application context when running
        # commands.
        context = None
        if self.app is not None:
            context = self.app.app_context()
            context.push()

        # Run as normal.
        result = super().run(*args, **kwargs)

        # Release the context.
        if context is not None:
            context.pop

        # Return the result.
        return result


def run(command=None):
    """Simple progammatic execution of the manager."""
    return Manager().handle(sys.argv[0], command or sys.argv[:1])



# -*- coding: utf-8 -*-
from alchemist import management


#! Global application context.
_app_context = None


def pytest_configure(config):
    # Configure the project environment; using utilties made available
    # by alchemist.

    # Discover the application that is being tested against.
    application = management.discover()

    # Establish an application context.
    global _app_context
    _app_context = application.app_context()
    _app_context.push()


def pytest_unconfigure(config):
    # Release the application context.
    _app_context.pop()



# -*- coding: utf-8 -*-
import io
import ipaddress


def build_database_uri(testing=False, **kwargs):
    # Resolve the expanded form into a database URI.
    o = io.StringIO()

    # Build key names and get properly.
    def key(name):
        if testing and ('test_' + name) in kwargs:
            return 'test_' + name

        return name

    # Add the designated engine.
    o.write(kwargs[key('engine')])
    o.write('://')

    # Add the username.
    user = False
    if key('user') in kwargs:
        o.write(kwargs[key('user')])
        user = True

    elif key('username') in kwargs:
        o.write(kwargs[key('username')])
        user = True

    # Add the password.
    if user:
        if key('pass') in kwargs:
            o.write(':')
            o.write(kwargs[key('pass')])

        elif key('password') in kwargs:
            o.write(':')
            o.write(kwargs[key('password')])

    # Add the hostname.
    null = 'localhost' if user else None
    hostname = kwargs.get(key('host'), kwargs.get(key('hostname'), null))
    if hostname:
        if user:
            # Write the user / host separator.
            o.write('@')

        ipaddr = True
        try:
            # Check if this is an ip address.
            ipaddress.ip_address(hostname)
            o.write('[')

        except ValueError:
            # This is not.
            ipaddr = False

        # Write out the hostname.
        o.write(hostname)
        if ipaddr:
            o.write(']')

    # Add the port.
    if key('port') in kwargs:
        o.write(':')
        o.write(str(kwargs[key('port')]))

    # Add the name.
    o.write('/')

    # Default a testing name if we have one.
    name = None
    if testing and 'test_name' not in kwargs:
        if kwargs[key('engine')] == 'sqlite':
            name = ':memory:'

        else:
            name = 'test_' + kwargs['name']

    if name is None:
        name = kwargs[key('name')]

    o.write(name)

    # Return our constructed URI.
    return o.getvalue()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os import path
from pkgutil import get_importer

# Calculate the base directory of the project to get relatives from.
BASE_DIR = path.abspath(path.dirname(__file__))

# Navigate, import, and retrieve the metadata of the project.
_imp = get_importer(path.join(BASE_DIR, 'src', 'alchemist'))
meta = _imp.find_module('meta').load_module('meta')

setup(
    name='alchemist',
    version=meta.version,
    description=meta.description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3'
    ],
    author='Concordus Applications',
    author_email='support@concordusapps.com',
    url='http://github.com/concordusapps/alchemist',
    scripts=['bin/alchemist'],
    package_dir={'alchemist': 'src/alchemist'},
    packages=find_packages(path.join(BASE_DIR, 'src')),
    entry_points={'pytest11': ['alchemist = alchemist.plugin']},
    dependency_links=(
        # wsgi_intercept
        # <https://code.google.com/p/wsgi-intercept/issues/detail?id=25>
        'git+git://github.com/concordusapps/wsgi-intercept.git'
        '#egg=wsgi_intercept-0.6.0',
    ),
    install_requires=(
        # Cross-platform colored terminal text.
        # <https://pypi.python.org/pypi/colorama>
        'colorama',

        # ANSII Color formatting for output in terminal.
        # <https://pypi.python.org/pypi/termcolor>
        'termcolor >= 1.1, < 1.2',

        # Flask is a microframework for Python based on Werkzeug,
        # Jinja 2 and good intentions.
        # <http://flask.pocoo.org/>
        'flask >= 0.10, < 0.11',

        # The Flask-Script extension provides support for writing external
        # scripts in Flask.
        # <http://flask-script.readthedocs.org/en/latest/>
        'flask-script >= 0.6.2',

        # The Python SQL Toolkit and Object Relational Mapper
        # <http://www.sqlalchemy.org/>
        'sqlalchemy >= 0.8, < 0.9',

        # SQLAlchemy utilities.
        # <https://github.com/kvesteri/sqlzalchemy-utils>.
        'sqlalchemy-utils',

        # Test runner.
        'pytest',

        # Ensure PEP8 conformance.
        'pytest-pep8',

        # Ensure test coverage.
        'pytest-cov',

        # Installs a WSGI application that intercepts requests made to a
        # hostname and port combination for testing.
        'wsgi_intercept == 0.6.0',

        # A comprehensive HTTP client library that supports many features
        # left out of other HTTP libraries.
        # TODO: This is required by wsgi_intercept; either the requirement
        # needs to removed (preferred) or it needs to be folded into
        # wsgi_intercept.
        'httplib2',

        # Requests.
        # Requests takes all of the work out of Python HTTP/1.1 —
        # making your integration with web services seamless.
        # <http://docs.python-requests.org/en/latest/>
        'requests'
    ),
)



# -*- coding: utf-8 -*-
import types
import sys


class Module(types.ModuleType):

    def __init__(self, *args, **kwargs):
        from sqlalchemy import orm

        # Initialize our module first.
        super().__init__(*args, **kwargs)

        # Create a scope function that keys on the application context.
        def scopefunc():
            import flask
            return id(flask._app_ctx_stack.top)

        # Create the scoped session factory which automatically creates
        # sessions when it needs to when accessed.
        self.Session = orm.scoped_session(
            self.create_session, scopefunc=scopefunc)

    @property
    def session(self):
        # Return the available session which is created automatically for
        # the current context if it does not exist.
        return self.Session()

    @property
    def create_session(self):
        # Save these for reference.
        from sqlalchemy import orm

        # Construct an inner class to late-bind configuration.
        class Session(orm.Session):

            def __init__(self, *args, **kwargs):
                from alchemist import application
                from alchemist.db.query import Query

                # Default the bind and query class.
                kwargs.setdefault('bind', application.databases['default'])
                kwargs.setdefault('query_cls', Query)

                # Continue the initialization.
                super().__init__(*args, **kwargs)

            def __repr__(self):
                return '<alchemist.db.Session(bind=%r)>' % self.bind

        # Return the inner class.
        return Session

    # Create helper proxies to the scoped session.

    def commit(self):
        return self.session.commit()

    def rollback(self, *args, **kwargs):
        return self.session.rollback(*args, **kwargs)

    def add(self, *targets):
        """Add the passed targets to the local session."""
        return self.session.add_all(targets)

    def remove(self):
        """Remove the local reference to the session."""
        return self.Session.remove()


# Update the inner module with the actual contents.
instance = Module(__name__)
instance.__dict__.update(sys.modules[__name__].__dict__)

# Store the module wrapper
sys.modules[__name__] = instance
